Module: ActiveItem::Logging
- Included in:
- Base
- Defined in:
- lib/active_item/logging.rb
Overview
Provides a dynamo_logger helper that delegates to the globally configured ActiveItem logger.
Provides a dynamo_logger helper that delegates to the globally configured ActiveItem logger.